# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# (c) 2007, 2008 Red Hat. Written by skvidal@fedoraproject.org, wwoods@redhat.com

# TODO: gpgcheck downloaded pkgs
# TODO: probably have some defensive checks before we start anything
#       system sanity checks, depsolver checks against current rpmdb
# TODO: deal with conflicts and try to walk around them
# TODO: generate local metadata

import sys
import os
import os.path
import shutil
from ConfigParser import ConfigParser
import xf86config
import shlex
import tempfile
from urlgrabber.grabber import URLGrabber, URLGrabError

import yum
import yum.Errors
import rpmUtils.arch

from yum.parser import varReplace
from yum.constants import *


__version__ = "0.9.3"

# this needs to be in config file somewhere
# we need a release identifier and baseurls/mirrorlists
# that's really it - maybe a gpgkey path....
if os.path.exists("releases.list"):
    releasesfn = "releases.list"
elif os.path.exists("data/releases.list"):
    releasesfn = "data/releases.list"
else:
    releasesfn = "/usr/share/preupgrade/releases.list"


def isParaVirt():
    """
    This function determines if the running system is a paravirt guest.
    This code is borrowed heavily from rhnreg.py (packaged with rhn-setup).
       2.  Check /sys/hypervisor/uuid.  If exists and is non-zero, we know
           the system is a para-virt guest; exit.
    """

    isGuest = False
    if os.path.exists("/proc/xen/xsd_port"):
        return False
    if os.path.exists("/sys/hypervisor/uuid"):
        try:
            uuid_file = open('/sys/hypervisor/uuid', 'r')
            uuid = uuid_file.read()
            uuid_file.close()
            if len(uuid.strip("0-\n")) != 0:
                isGuest = True
        except IOError:
            # Failed; must not be para-virt.
            pass

    return isGuest

def getlang():
    lang='en_US' # fallback
    infile = '/etc/sysconfig/i18n'
    s = shlex.shlex(open(infile),infile,posix=True)
    while True:
        t = s.get_token()
        if not t:
            break
        if t == 'LANG':
            eq = s.get_token() # =
            if eq == '=':
                lang = s.get_token()  # value
    return lang

def st_dev_to_devicename(st_dev):
    '''Try to convert st_dev to a devicename by walking /sys/block.
    Yes, this will only work for block devices. Yes, that's intentional.'''


    (major,minor) = (os.major(st_dev),os.minor(st_dev))


    def prune_func(rightone, dirname, filenames):
        if rightone:
            return
        if dirname == '/sys/block':
            return

        if 'dev' in filenames:
            dev_f = open(dirname+'/dev')
            dev = dev_f.readline().strip()
            dev_f.close()
        else:
            return
        if dev == '%s:%s' % (major, minor):
            rightone.append(os.path.basename(dirname))
            return

    places_to_look = []
    for r,d,f in os.walk('/sys/block'):
        for thisdir in d:
            places_to_look.append(r+'/'+thisdir)

    rightone = []
    for place in places_to_look:
        os.path.walk(place, prune_func, rightone)
        if rightone:
            return rightone[0]
    
    return None

def get_mdraidinfo(devname):
    '''If devname is a mdraid device, return a dict with the following keys:
    {'level':"RAID level", 'members':[list of active, in-sync member devices]}. Example:
    {'level':'raid1','members':['sda1','sdb1]}
    Otherwise, returns an empty dict.'''
    results = {}
    md_sysdir="/sys/block/%s/md/" % devname

    # Get the RAID level; if this fails we don't have a proper md dir
    # (and thus this probably isn't a RAID device at all)
    try:
        f = open(md_sysdir+"level")
    except IOError:
        return results
    results['level'] = f.readline().strip()
    f.close()
    # Yep, it's RAID. Let's read in the member devices.
    # Info about how this works comes from md.txt in kernel-docs
    n = 0
    rd = md_sysdir+"rd%i" % n
    members = []
    while os.path.exists(rd):
        dev = os.readlink(rd).replace("dev-","",1)
        members.append(dev)
        n += 1
        rd = md_sysdir+"rd%i" % n
    results['members'] = members
    return results

def bootpath_to_anacondapath(somepath):
    '''Convert a path under /boot to a path that anaconda will understand.'''
    st_dev = os.stat(somepath).st_dev
    bootdev = st_dev_to_devicename(st_dev)
    mtab = open("/etc/mtab")
    mountpoint = ''
    for line in mtab:
        (dev,mp,rest) = line.split(' ',2)

        if dev.endswith(bootdev):
            mountpoint = mp
    raid = get_mdraidinfo(bootdev)
    if raid:
        if raid['level'] == 'raid1':
            bootdev = raid['members'][0]
        else:
            print "Huh? /boot is unbootable RAID level '%s'" % raid['level']
    if mountpoint:
        return (bootdev,somepath.replace(mountpoint,'',1))

def getkeymap():
    '''Get the system keymap from xorg.conf or /etc/sysconfig/keyboard'''
    layout='us' # fallback
    # First, try xorg.conf
    (xconfig, xconfigpath) = xf86config.readConfigFile()
    if xconfig:
        try:
            keyboard = xf86config.getCoreKeyboard(xconfig)
        except xf86config.XF86SectionMissing, e:
            print 'Warning getting keymap: could not find keyboard section'
            return layout
        
        for o in keyboard.options:
            if o.name == 'XkbLayout':
                layout = o.val
        if '+' in layout:
            layout = layout.split('+')[0]
    # No xorg.conf - try /etc/sysconfig/keyboard
    else:
        infile = '/etc/sysconfig/keyboard'
        try:
            s = shlex.shlex(open(infile),infile,posix=True)
        except IOError:
            pass
        while s:
            t = s.get_token()
            if not t:
                break
            if t == 'KEYTABLE':
                eq = s.get_token() # =
                if eq == '=':
                    layout = s.get_token()  # value
    return layout
    
def grubby(kernel,initrd,title="upgrade fedora",args='',):
    # FIXME check for (and overwrite) existing boot entry
    cmd='/sbin/grubby --title="%s"' % title
    if args != '':
        cmd=cmd+' --add-kernel=%s --initrd=%s --args="%s"' % (kernel,initrd,args)
    else:
        cmd=cmd+' --add-kernel=%s --initrd=%s' % (kernel,initrd)
    #DEBUG FIXME
    print 'DEBUG %s' % cmd
 
    return os.system(cmd)
    
def grubby_make_next_boot_only():
    return os.system("/bin/echo 'savedefault --default=0 --once' | /sbin/grub >/dev/null")
    



class PUError(Exception):
    def __init__(self, value=None):
        Exception.__init__(self)
        self.value = value
    def __str__(self):
        return "%s" %(self.value,)

class PreUpgrade(yum.YumBase):
    
    def __init__(self, release=None, configfile=releasesfn):
        yum.YumBase.__init__(self)
        self.release_map = self.retrieve_release_info(configfile)
        self.bootpath='/boot/upgrade'
        if release:
            self.setup(release)

    def upgrade_available(self, stable=True):
        '''If there is a newer release available in the release map, return its
        name/key. If there are multiple newer releases, it returns the newest
        one (the one with the highest version number).
        Otherwise returns None.
        You can set stable to False if you want to consider non-stable releases
        as well (but right now this always gets you Rawhide).'''
        current = self.conf.yumvar['releasever']
        target_ver = current
        target_key = None
        for release_key in self.release_map:
            ver = self.release_map[release_key]['version']
            release_stable = 'stable' in self.release_map[release_key]
            if stable and not release_stable:
                continue
            # FIXME: unless we return a list or something, we'll always return 
            # Rawhide if stable=False, since it's always available and always
            # considered newer than whatever you've got.
            if ver > target_ver: 
                target_ver = ver
                target_key = release_key
        return target_key

    def setup(self, release_key):
        if self.release_map.has_key(release_key):
            rel = self.release_map[release_key]
            self.pu_release = release_key

            self.pu_mirrorlist        = rel.get('mirrorlist')
            self.pu_baseurl           = rel.get('baseurl')
            self.pu_installmirrorlist = rel.get('installmirrorlist')
            self.pu_installurl        = rel.get('installurl')
            if not (self.pu_installmirrorlist or self.pu_installurl):
                self.pu_installmirrorlist = self.pu_mirrorlist
                self.pu_installurl = self.pu_baseurl
        else:
            raise PUError, 'Release does not match any in configuration.'

    def retrieve_release_info(self, configfile=None):
        releases = {}
        cp = ConfigParser()
        cp.read(configfile)
        for sect in cp.sections():
            rel = dict(cp.items(sect))
            if ('baseurl' in rel) or ('mirrorlist' in rel):
                releases[sect] = rel
                
        return releases
    
    def release_is_stable(self,release):
        '''Return True if the given release is a stable release'''
        # this is an abstraction - just in case we change the structure of the
        # 'release' data.
        stable = release.get('stable')
        if type(stable) == type(''):
            stable = stable.lower()
            if stable == '1' or stable == 'true' or stable == 'yes':
                return True
        # else
        return False
        
    def setup_update_repo(self, download_progressbar=None):
        # disable all the other repos
        self.repos.disableRepo('*')
        # make our new repo obj
        newrepo = yum.yumRepo.YumRepository('anaconda-upgrade')
        newrepo.name = 'Anaconda Upgrade - %s' % self.pu_release
        if self.pu_mirrorlist: 
            self.pu_mirrorlist = varReplace(self.pu_mirrorlist, self.conf.yumvar)
        newrepo.mirrorlist = self.pu_mirrorlist
        newrepo.baseurl = []
        if self.pu_baseurl: 
            self.pu_baseurl = varReplace(self.pu_baseurl, self.conf.yumvar)
            newrepo.baseurl = self.pu_baseurl
        newrepo.enablegroups = 1
        
        # Set up a second yumRepo for boot images etc.)
        # (this is used by _retrieve_file)
        instrepo = yum.yumRepo.YumRepository('anaconda-upgrade-install')
        instrepo.name = 'Anaconda Upgrade (installable tree) - %s' % self.pu_release
        if self.pu_installmirrorlist:
            self.pu_installmirrorlist = varReplace(self.pu_installmirrorlist,self.conf.yumvar)
        instrepo.mirrorlist = self.pu_installmirrorlist
        instrepo.baseurl = []
        if self.pu_installurl:
            self.pu_installurl = varReplace(self.pu_installurl,self.conf.yumvar)
            instrepo.baseurl = self.pu_installurl

        if download_progressbar:
            instrepo.setCallback(download_progressbar)

        # Save this secondary repo object
        self.instrepo = instrepo
        # This is a little dicey, using private stuff from yumRepo. Oh well.
        self.instgrab = self.instrepo._getgrab()
        
        newrepo.basecachedir = '/var/cache/yum/'
        instrepo.basecachedir = '/var/cache/yum/'

        # add our new repo
        self.repos.add(newrepo)        
        # enable that repo
        self.repos.enableRepo(newrepo.id)
        # setup the progress bars for this repo        
        if download_progressbar:
            newrepo.setCallback(download_progressbar)
        # setup the repo dirs/etc
        self.doRepoSetup(thisrepo=newrepo.id)
        self.distrepo = self.repos.findRepos(newrepo.id)[0]
        # turn on caching
        self.conf.keepcache = 1

    
    def retrieve_treeinfo(self):
        # take our repo
        # find a .treeinfo in them
        # download it
        # parse it
        # store relative path to the right kernel and initrd
        # there can be only one
        # info we need:
        # initrd.img path
        # kernel path
        # stage2.img path
        # min image
        # example:
        #   http://download.fedora.redhat.com/pub/fedora/linux/development/i386/os/.treeinfo
        tif = self.instrepo.grab.urlopen('/.treeinfo')
        #cracked out ppc issues, etc
        archlist = [self.conf.yumvar['basearch']]
        if isParaVirt():
            archlist = ['xen']
        if rpmUtils.arch.getCanonArch().startswith('ppc64'):
            archlist = [rpmUtils.arch.getCanonArch(), self.conf.yumvar['basearch']]
        
        cp = ConfigParser()
        cp.readfp(tif)
        
        mysection = None
        for arch in archlist:
            if cp.has_section('images-%s' % arch):
                mysection = 'images-%s' % arch
                break
        
        if not mysection:
            raise PUError, "Cannot find any kernel/initrd section for this arch in treeinfo"
            
        self.kernel = cp.get(mysection, 'kernel')
        self.initrd = cp.get(mysection, 'initrd')
        if cp.has_section('stage2'):
            self.instimage = cp.get('stage2', 'instimage')
            self.mainimage = cp.get('stage2', 'mainimage')
        else:
            self.instimage = 'images/minstg2.img'
            self.mainimage = 'images/stage2.img'

        tif.close()
        
    
    def _retrieve_file(self, fileinfo, targetdir=''):
        """ for returning files from our important repo, but not package or metadata files"""
        if not targetdir:
            targetdir = self.bootpath
        
        if not os.path.exists(targetdir):
            os.makedirs(targetdir)

        # for each file:
        # check if we have the filesize and check it out
        # if it is okay, then continue
        # if it is not okay then unlink it 
        # check for space to download it
        
        # local name
        item_fname = os.path.basename(fileinfo)
        local = targetdir + '/' + item_fname
        # remote size
        tmp = self.instgrab.urlopen(fileinfo)
        mysize = tmp.hdr.dict['content-length']
        tmp.close()

        # see if we have it and if it is the right size        
        if os.path.exists(local):
            local_stat = os.stat(local)
            if int(local_stat.st_size) == int(mysize):
                # File exists and it's the right size.. guess it's probably OK
                # We should be doing some integrity checks but we don't have
                # anything to check it against - la la la la
                return True
            else:
                os.unlink(local)

        # if we are this far we're going to try to download it
        # check for available disk space
        dirstat = os.statvfs(targetdir)
        if (dirstat.f_bavail * dirstat.f_bsize) <= long(mysize):
            raise PUError,("Not enough space in %s to cache %s.\n\n" +
                          "It will be downloaded once the installer starts.")\
                           % (targetdir, item_fname)

        try:
            self.instrepo._getFile(relative=fileinfo, local=local)
        except URLGrabError, e:
            os.unlink(local)
            raise PUError, "failure downloading %s: %s" % (item_fname, e)
            
        return True

    def remove_boot_files(self):
        '''Remove any previously-downloaded boot files.'''
        def rmtree_handler(func,path,exc_info):
            (e_type, e_value, e_traceback) = exc_info
            raise PUError, "Failed to remove %s: %s" % (path,e_value)
        if os.path.exists(self.bootpath):
            shutil.rmtree(self.bootpath,ignore_errors=False,\
                          onerror=rmtree_handler)

    def remove_repo_cache(self):
        '''Remove the repo caches, if they exist.
        Will throw PUError if the repos are not yet set up.'''
        # Since we inherit from YumBase we've got the cleanXXX methods:
        #   self.clean{Headers,Packages,Sqlite,Metadata}
        if not (hasattr(self,'distrepo') or hasattr(self,'instrepo')):
            # we need to set up the repos before we can clean them!
            raise PUError, "Can't clear repo caches until repos are set up"
        self.cleanHeaders()
        self.cleanPackages()
        self.cleanSqlite()
        self.cleanMetadata()
        
    def resuming_run(self):
        '''If there is cached data from a previously-interrupted run of
        preupgrade, return the release name (self.pu_release)'''
        # FIXME implement!
        print "PreUpgrade.resuming_run(): STUB!"
        return ""

    def retrieve_critical_boot_files(self):
        # relative tmp FIXME XXX
        
        for item in (self.kernel, self.initrd):
            self._retrieve_file(item)
        
    def retrieve_non_critical_files(self,targetdir=''):
        self._retrieve_file(self.mainimage,targetdir)
        return os.path.basename(self.mainimage)
        
    def get_download_pkgs(self):
       return map(lambda txmbr: txmbr.po,
               self.tsInfo.getMembersWithState(output_states=TS_INSTALL_STATES))

    def return_available_versions(self, cur_ver, beta_ok=False):
        versions = []
        for name, releasedata in self.release_map.iteritems():
            # Skip releases that aren't newer than this one
            if cur_ver >= float(releasedata.get('version')):
                continue
            # If beta_ok get all releases, otherwise just the stable ones
            if beta_ok or self.release_is_stable(releasedata):
                versions.append(name)
        return versions
        
    def figure_out_what_we_need(self):
        # grab the comps file, we'll need it later for anaconda
        if self.comps.compscount == 0:
            raise PUError, "Cannot retrieve group lists from any repository"
        # also try to get the gzipped comps file if it's there so that
        # anaconda doesn't have to later.  this is kind of ugly (#439232)
        for r in self.repos.listGroupsEnabled():
            try:
                r.retrieveMD('group_gz')
            except:
                pass

        # update all
        self.update()
        # try and remove all the old kernels
        self.remove(name='kernel')
        self.remove(name='kernel-xen')
        # build a transaction
        self.buildTransaction()
        # print out what we would be doing

    def make_it_boot(self,extra_args=''):
        mykernel  = self.bootpath  + '/' + os.path.basename(self.kernel)
        myinitrd = self.bootpath + '/' + os.path.basename(self.initrd)
        lang = getlang()
        keymap = getkeymap()
        arch = self.conf.yumvar['basearch']
        argstr = "preupgrade lang=%s keymap=%s" % (lang,keymap)
        if extra_args: 
            argstr += " " + extra_args
        title="Upgrade to %s" % self.pu_release
        # yaboot (ppc) doesn't support spaces in titles
        if arch.startswith('ppc'):
            title = title.replace(' ','_')
        #FIXME - do something with return codes below here :)

        grubby(mykernel, myinitrd, title=title, args=argstr)
        grubby_make_next_boot_only()
        # ppc machines need to run ybin to activate changes
        if arch.startswith('ppc'):
            os.system('/sbin/ybin > /dev/null')
            
    def _main(self):
        '''Basically an example of how a run is supposed to proceed.'''
        self.setup('Rawhide')
        self.setup_update_repo()        
        # In this example, we always clear old data rather than reusing it
        self.remove_repo_cache()
        self.remove_boot_files()
        self.figure_out_what_we_need()
        downloadpkgs = []
        downloadpkgs = map(lambda txmbr: txmbr.po,
                           self.tsInfo.getMembersWithState(output_states=TS_INSTALL_STATES))
        del(self.tsInfo)
        
       
        problems = self.downloadPkgs(downloadpkgs) 

        # find tree info data 
        self.retrieve_treeinfo()
        
        # grab kernel and initrd and stage2
        try:
            self.retrieve_critical_boot_files()
        except PUError, e:
            print 'Fatal: %s' % e
            sys.exit(1)

        try:            
            self.retrieve_non_critical_boot_files()        
        except PUError:
            pass
        
        # make it bootable
        self.make_it_boot()        


